import 'dart:ffi';
import 'dart:typed_data';
import 'package:ffi/ffi.dart';
import 'package:ffi/ffi.dart' as ffi;
import 'package:meta/meta.dart';
import 'array_utils.dart';

/// Represents a [Pointer] and its length.
class Array<T extends NativeType> {
  @doNotStore
  final Pointer<T> ptr;
  final length;
  TypedData _typedList;

  Array.fromPtr(this.ptr, this.length);

  Array(this.length, [Allocator allocate = ffi.allocate])
      : ptr = allocate<T>(count: length);

  /// Frees [ptr] with [ffi.free]. Override this method if you need
  /// to perform actions before freeing or to use a different deallocator.
  void delete() {
    free(ptr);
  }

  @override
  bool operator ==(dynamic arr) {
    if (arr is Pointer) {
      return arr == ptr;
    }
    if (arr is Array) {
      return arr.ptr == ptr;
    }
    return false;
  }

  @override
  int get hashCode => ptr.hashCode;

  @override
  String toString() {
    return '${runtimeType}($ptr, $length)';
  }
}

/// [ArrayView] is a non-owning view into allocated memory,
/// meaning [delete] doesn't do anything.
class ArrayView<T extends NativeType> extends Array<T> {
  ArrayView.fromPtr(Pointer<T> ptr, int length) : super.fromPtr(ptr, length);
  ArrayView(int length, [Allocator allocate]) : super(length, allocate);
  @override
  void delete() {}
}

extension Uint8Array on Array<Uint8> {
  @doNotStore
  Uint8List asTypedList() {
    _typedList ??= ptr.asTypedList(length);
    return _typedList;
  }
}
